; Definitions for combinator graphs.  Intended for package *LISP

(defvar SYS-mode 'BUILDING)    ; system mode

(*proclaim '(type fixnum CA-SZ))
(DEFCONSTANT CA-SZ 18 "Size of address fields") ; allows 256k v procs

(*PROCLAIM '(TYPE (FIELD-PVAR CA-SZ) LPTR LPTX RPTR RPTX CXTP CXTX CPYP RV))

(*DEFVAR LPTR (!! 0) "Left descendant")

(*DEFVAR LPTX (!! 0) "Next Left descendant")

(*DEFVAR RPTR (!! 0) "Right descendant")

(*DEFVAR RPTX (!! 0) "Next Right descendant")

(*DEFVAR CXTP (!! 0) "Expression context")

(*DEFVAR CXTX (!! 0) "Next expression context")

(*DEFVAR CPYP (!! 0) "Copy pointer")

(*DEFVAR RV (!! 0) "Rendezvous pointer")

(*PROCLAIM '(TYPE (SIGNED-PVAR (1+ CA-SZ)) REFC QNTC))

(*DEFVAR REFC (!! 0) "Reference counts")

(*DEFVAR QNTC (!! 0) "Quantifier counts")

(DEFCONSTANT TY-SZ 8 "Size of type field")  ; allows 256 types

(*PROCLAIM '(TYPE (FIELD-PVAR TY-SZ) GTYP RDTN))

(*DEFVAR GTYP (!! 0) "Types")

(*DEFVAR GTYX (!! 0) "New Types")

(*DEFVAR RDTN (!! 0) "Reduction")

(*PROCLAIM '(TYPE SIGNED-PVAR XNMB))

(*DEFVAR XNMB (!! 0) "Numerical value, when needed")

(*PROCLAIM
 '(TYPE BOOLEAN-PVAR NRMT NMLT NMRT STRT STXT VIST ENVT PCXT QNTT IRRT))

(*DEFVAR NRMT nil!! "Normalized expression")

(*DEFVAR NMLT nil!! "L is normal")

(*DEFVAR NMRT nil!! "R is normal")

(*DEFVAR STRT nil!! "Structure")

(*DEFVAR STXT nil!! "Next structure")

(*DEFVAR VIST nil!! "Visited")

(*DEFVAR ENVT nil!! "In current symbol env")

(*DEFVAR PCXT nil!!)  ; used to manage context propagation in and's

(*DEFVAR QNTT nil!!)  ; node is quantified

(*DEFVAR IRRT nil!!)  ; node is irreducible



(proclaim '(type fixnum root))
(defvar ROOT 1)

(defvar dictionary nil)               ; atoms introduced by user

; Access macros for principal pvars

(defmacro L (i) "L[i]" `(pref LPTR ,i))
(defmacro LX (i) "LX[i]" `(pref LPTX ,i))
(defmacro R (i) "R[i]" `(pref RPTR ,i))
(defmacro RX (i) "RX[i]" `(pref RPTX ,i))
(defmacro C (i) "C[i]" `(pref CXTP ,i))
(defmacro CX (i) "CX[i]" `(pref CXTX ,i))
(defmacro P (i) "P[i]" `(pref CPYP ,i))
(defmacro F (i) "REF[i]" `(pref REFC ,i))
(defmacro Q (i) "QNT[i]" `(pref QNTC ,i))
(defmacro G (i) "G[i]" `(pref GTYP ,i))
(defmacro GX (i) "GX[i]" `(pref GTYX ,i))
(defmacro RN (i) "RN[i]" `(pref RDTN ,i))
(defmacro X (i) "XNMB[i]" `(pref XNMB ,i))
(defmacro N (i) "NRM[i]" `(pref NRMT ,i))
(defmacro NL (i) "NMLT[i]" `(pref NMLT ,i))
(defmacro NR (i) "NMLT[i]" `(pref NMRT ,i))
(defmacro S (i) "STR[i]" `(pref STRT ,i))
(defmacro STX (i) "STX[i]" `(pref STXT ,i))
(defmacro V (i) "VIS[i]" `(pref VIST ,i))
(defmacro E (i) "ENV[i]" `(pref ENVT ,i))
(defmacro QNT (i) "QNTT[i]" `(pref QNTT ,i))
(defmacro RV (i) "RV[i]" `(pref RV ,i))

(defmacro L!! (i) "L!![i]" `(pref!! LPTR ,i))
(defmacro LX!! (i) "LX!![i]" `(pref!! LPTX ,i))
(defmacro R!! (i) "R!![i]" `(pref!! RPTR ,i))
(defmacro RX!! (i) "RX!![i]" `(pref!! RPTX ,i))
(defmacro C!! (i) "C!![i]" `(pref!! CXTP ,i))
(defmacro CX!! (i) "CX!![i]" `(pref!! CXTX ,i))
(defmacro P!! (i) "P!![i]" `(pref!! CPYP ,i))
(defmacro F!! (i) "REF!![i]" `(pref!! REFC ,i))
(defmacro Q!! (i) "QNT!![i]" `(pref!! QNTC ,i))
(defmacro G!! (i) "G!![i]" `(pref!! GTYP ,i))
(defmacro GX!! (i) "GX!![i]" `(pref!! GTYX ,i))
(defmacro RN!! (i) "RN!![i]" `(pref!! RDTN ,i))
(defmacro X!! (i) "XNMB!![i]" `(pref!! XNMB ,i))
(defmacro NL!! (i) "NMLT!![i]" `(pref!! NMLT ,i))
(defmacro NR!! (i) "NMRT!![i]" `(pref!! NMRT ,i))
(defmacro N!! (i) "NRM!![i]" `(pref!! NRMT ,i))
(defmacro S!! (i) "STR!![i]" `(pref!! STRT ,i))
(defmacro STX!! (i) "STX!![i]" `(pref!! STXT ,i))
(defmacro V!! (i) "VIS!![i]" `(pref!! VIST ,i))
(defmacro E!! (i) "ENV!![i]" `(pref!! ENVT ,i))
(defmacro QNT!! (i) "QNT!![i]" `(pref!! QNTT ,i))
(defmacro RV!! (i) "RV!![i]" `(pref!! RV ,i))

; Macros defining type codes.  All are unsigned, 8-bit integers
; Special node types:    128 - L is valid ptr,  64 - R is valid ptr
; 32 - C is a valid.  Note that a 0 pointer is a dummy
(defmacro ty-free () "Free node" 0)
(defmacro ty-numb () "Numeral" 1)
(defmacro ty-comb () "Combination" (+ 128 64 32 2))
(defmacro ty-cons () "Construction" (+ 128 64 3))
(defmacro ty-var () "Variable" (+ 64 32 4))
(defmacro ty-point () "Pointer" (+ 64 5))
(defmacro ty-dict () "Dictionary atom" 6)
(defmacro ty-sysc () "System constant" 7)
(defmacro ty-notn () "Notion" (+ 64 8))
(defmacro ty-dummy () "Dummy node" 9)
(defmacro ty-new () "New node" 10)
(defmacro ty-dead () "Dead node" 15)

; Tests for valid pointers

(defmacro VLP!! () "(VLP!!) - L is valid"
 '(and!! (plusp!! (load-byte!! GTYP (!! 7) (!! 1))) (plusp!! LPTR))
)

(defmacro VRP!! () "(VRP!!) - R is valid"
 '(and!! (plusp!! (load-byte!! GTYP (!! 6) (!! 1))) (plusp!! RPTR))
)

(defmacro VCP!! () "(VCP!!) - C is valid and counted"
 '(and!! (VARP!!) (plusp!! CXTP))
)

; Parallel type tests

(defmacro FREEP!! () '(=!! GTYP (!! (ty-free))))

(defmacro NUMBP!! () '(=!! GTYP (!! (ty-numb))))

(defmacro COMBP!! () '(=!! GTYP (!! (ty-comb))))

(defmacro CONSP!! () '(=!! GTYP (!! (ty-cons))))

(defmacro VARP!! () '(=!! GTYP (!! (ty-var))))

(defmacro POINTP!! () '(=!! GTYP (!! (ty-point))))

(defmacro DICTP!! () '(=!! GTYP (!! (ty-dict))))

(defmacro SYSCP!! () '(=!! GTYP (!! (ty-sysc))))

(defmacro NOTNP!! () '(=!! GTYP (!! (ty-notn))))

(defmacro DMMYP!! () '(=!! GTYP (!! (ty-dummy))))

(defmacro NEWP!! () '(=!! GTYP (!! (ty-new))))

(defmacro DEADP!! () (=!! GTYP (!! (TY-dead))))

; Symbol table - maps symbol code to symbols

(defvar symnum 0 "Number of symbols encountered")

(defvar symvec NIL "Symbol array, indexed by name-code")


(defun init-sym ()
"(init-sym) -- resets symbol def'ns and initializes array"
 (cond (symvec (map NIL #'(lambda (x) (remprop x 'CMSYM)) symvec)
               (setf (fill-pointer symvec) 0))
       (T (setq symvec (make-array '(1024) :initial-element 0 :adjustable T
                                   :fill-pointer 0))))
 (setq symnum 0)
)

(init-sym)

(defun entersym (id)
"(entersym id) - Enter id in symbol table"
 (prog1 symnum
  (vector-push-extend id symvec 512)
  (setf (get id 'CMSYM) symnum)
  (incf symnum))
)

; Built-in operators,  0 <= <op-code> < syscmx.
; "Barred" combinators have codes 64 + (unbarred comb)
; Cube address is now same as type-code, so macros have been eliminated.

(defun initty (id ty) "initialize operator id with type ty"
 (sC ty ty) (sG ty (ty-sysc)) (sX ty (get id 'CMSYM)) (sN ty T) (sE ty T))

(defun inixty (id cb ty)
 (setf (get id 'OPERATOR) (cons ty cb)) (entersym id)
 (setf (aref optyp ty) id) (set cb ty) (push id oplst))

(defmacro inity (id cb ty) "operator initialization"
 `(progn (inixty ',id ',cb ,cb)))

(defvar syscmx 70)                  ; Largest sysc code
(defvar oplst ())                   ; List of operator symbols
(defvar optyp (make-array (list (1+ syscmx))))
                                    ; Operator symbol is optyp[<op-code>]
(setq oplst NIL)                    ; Permits reloading

; Garbage collection frequency:

(defvar GBGCYCLE 1)   ; number of garbage collections per firing

(defvar CFREE 0)      ; number of nodes available for CCPV's

(defvar atml 0) (defvar ccpl 0) (defvar lfree 0) (defvar hfree 0)

(defun initstore() "(initstore) - initializes storage management"
 (setq ATML (1+ syscmx)) (setq CCPL (1- cm:*user-cube-address-limit*))
 (setq LFREE (setq HFREE 0)) (setq CFREE (- CCPL ATML -1))
)

; inity calls below show the old macro, but it's ignored.

; I, K, K2 must have codes < 16,  S, B, C should be adjacent, S at 4

(defconstant CB-I 1)           (inity *I CB-I (ty-I))
(defconstant CB-K 2)           (inity *K CB-K (ty-K))
(defconstant CB-K2 3)          (inity *K2 CB-K2 (ty-K2))
(defconstant CB-S 4)           (inity *S CB-S (ty-S))
(defconstant CB-SB (+ 64 4))   (inity *SB CB-SB (ty-SB))
(defconstant CB-B 5)           (inity *B CB-B (ty-B))
(defconstant CB-BB (+ 64 5))   (inity *BB CB-BB (ty-BB))
(defconstant CB-C 6)           (inity *C CB-C (ty-C))
(defconstant CB-CB (+ 64 6))   (inity *CB CB-CB (ty-CB))
(defconstant CB-W  7)          (inity *W CB-W (ty-W))

; Head, tail should be adjacent, even-odd

(defconstant CB-head 8)        (inity hd CB-head (ty-head))
(defconstant CB-tail 9)        (inity tl CB-tail (ty-tail))
(defconstant CB-pair 10)       (inity pr  CB-pair  (ty-pair))
(defconstant CB-= 11)          (inity = CB-= (ty-equals))
(defconstant CB-Y 12)          (inity *Y CB-Y (ty-Y))
(defconstant CB-is 13)         (inity is CB-is (ty-is))

; false, true are an adjacent, even-odd pair  < 16

(defconstant CB-false 14)      (inity false CB-false (ty-false))
(defconstant CB-true 15)       (inity true  CB-true  (ty-true))

; Boolean ops have 16 <= code <= CB-if

(defconstant CB-and 16)        (inity and CB-and (ty-and))
(defconstant CB-or 17)         (inity or CB-or (ty-or))
(defconstant CB-not 18)        (inity not CB-not (ty-not))
(defconstant CB-if 19)         (inity if CB-if (ty-if))

(proclaim '(type fixnum maxbool))
(defvar maxbool (- CB-if 16))

; Sysc's which are structures should be between [] and inf.

(defconstant CB-[] 30)         (inity [] CB-[] (ty-[]))
(defconstant CB-inf 31)        (inity inf CB-inf (ty-inf))

; Arithmetic ops, all have 32 <= tyc < 64

(defconstant CB-+ 32)          (inity + CB-+ (ty-sum))
(defconstant CB-* 33)          (inity * CB-* (ty-product))
(defconstant CB-- 34)          (inity - CB-- (ty-difference))
(defconstant CB-/ 35)          (inity / CB-/ (ty-quotient))
(defconstant CB-< 36)          (inity <  CB-< (ty-less))
(defconstant CB-<= 37)         (inity <= CB-<= (ty-lesseql))
(defconstant CB-> 38)          (inity >  CB->  (ty-greater))
(defconstant CB->= 39)         (inity >= CB->= (ty-grteql))

;Table for and-or-not-if branches in reduction

(defvar and-ifcd 0 "value for table look-up")
(*proclaim '(type (field-pvar 32) AND-IF))
(*defvar AND-IF (!! 0) "Table")

;Element assignment:

(defmacro sL (i v)  "(sL i v)  -- L[i] := v" `(setf (pref LPTR ,i) ,v))

(defmacro sR (i v)  "(sR i v)  -- R[i] := v" `(setf (pref RPTR ,i) ,v))

(defmacro sC (i v)  "(sC i v)  -- C[i] := v" `(setf (pref CXTP ,i) ,v))

(defmacro sF (i v)  "(sF i v)  -- REF[i] := v" `(setf (pref REFC ,i) ,v))

(defmacro sG (i v)  "(sG i v)  -- G[i] := v" `(setf (pref GTYP ,i) ,v))

(defmacro sX (i v)  "(sX i v)  -- XNMB[i] := v"
 `(setf (pref XNMB ,i) ,v))

(defmacro sN (i v)  "(sN i v)  -- N[i] := v" `(setf (pref NRMT ,i) ,v))

(defmacro sNL (i v)  "(sNL i v)  -- NL[i] := v" `(setf (pref NMLT ,i) ,v))

(defmacro sNR (i v)  "(sNR i v)  -- NR[i] := v" `(setf (pref NMRT ,i) ,v))

(defmacro sS (i v)  "(sS i v)  -- S[i] := v" `(setf (pref STRT ,i) ,v))

(defmacro sSTX (i v)  "(sSX i v)  -- SX[i] := v" `(setf (pref STXT ,i) ,v))

(defmacro sV (i v)  "(sV i v)  -- V[i] := v" `(setf (pref VIST ,i) ,v))

(defmacro sE (i v)  "(sE i v)  -- E[i] := v" `(setf (pref ENVT ,i) ,v))

(defmacro sQNT (i v)  "(sQNT i v)  -- QNT[i] := v" `(setf (pref QNTT ,i) ,v))

;Parallel assignment:

(defmacro sL!! (v)  "(sL!! v)  -- L !!= v" `(*set LPTR ,v))

(defmacro sLX!! (v)  "(sLX!! v)  -- LX !!= v" `(*set LPTX ,v))

(defmacro sR!! (v)  "(sR!! v)  -- R !!= v" `(*set RPTR ,v))

(defmacro sRX!! (v)  "(sRX!! v)  -- RX !!= v" `(*set RPTX ,v))

(defmacro sC!! (v)  "(sC!! v)  -- C !!= v" `(*set CXTP ,v))

(defmacro sCX!! (v)  "(sCX!! v)  -- CX !!= v" `(*set CXTX ,v))

(defmacro sP!! (v)  "(sP!! v)  -- P !!= v" `(*set CPYP ,v))

(defmacro sF!! (v)  "(sF!! v) -- REF !!= v" `(*set REFC ,v))

(defmacro sQ!! (v)  "(sQ!! v) -- QNT !!= v" `(*set QNTC ,v))

(defmacro sG!! (v)  "(sG!! v)  -- G !!= v" `(*set GTYP ,v))

(defmacro sGX!! (v)  "(sGX!! v)  -- GX !!= v" `(*set GTYX ,v))

(defmacro sRN!! (v)  "(sRN!! v)  -- RN !!= v" `(*set RDTN ,v))

(defmacro sX!! (v)  "(sX!! v)  -- XNMB !!= v" `(*set XNMB ,v))

(defmacro sN!! (v)  "(sN!! v)  -- N !!= v" `(*set NRMT ,v))

(defmacro sNL!! (v)  "(sNL!! v)  -- NL !!= v" `(*set NMLT ,v))

(defmacro sNR!! (v)  "(sNR!! v)  -- NR !!= v" `(*set NMRT ,v))

(defmacro sS!! (v)  "(sS!! v)  -- S !!= v" `(*set STRT ,v))

(defmacro sSTX!! (v)  "(sSX!! v)  -- SX !!= v" `(*set STXT ,v))

(defmacro sV!! (v)  "(sV!! v)  -- V !!= v" `(*set VIST ,v))

(defmacro sE!! (v)  "(sE!! v)  -- E !!= v" `(*set ENVT ,v))

(defmacro sQNT!! (v)  "(sQNT!! v)  -- QNT !!= v" `(*set QNTT ,v))

(defmacro sRV!! (v)  "(sRV!! v)  -- RV !!= v" `(*set RV ,v))

; Helpful macro for common communication situation

(defmacro *psetno (s d c)
"(*psetno source dest cube-addr) -- *pset :no-collisions"
 `(*pset :no-collisions ,s ,d ,c)
)

; Reduction statistics:

(*PROCLAIM '(type (field-pvar 30) STATS))

(*DEFVAR STATS (!! 0))
(defvar *STATS* T)    ; controls collection of statistics

(defvar REDUCTIONS 0) ; records total number of reductions each firing
(defvar TOTRDTNS   0) ; records total reductions since inst[all]
(defvar FIRINGS    0) ; records number of calls to FIRE

; Logic scheme control:

(defvar SCHEME 'B)    ; selects among schemes A, B, C

; Working variables for FIRE:

(*PROCLAIM '(type (field-pvar CA-SZ) LL LLL A B C D E F G))

(*DEFVAR LL (!! 0))
(*DEFVAR LLL (!! 0))
(*DEFVAR A (!! 0))
(*DEFVAR B (!! 0))
(*DEFVAR C (!! 0))
(*DEFVAR D (!! 0))
(*DEFVAR E (!! 0))
(*DEFVAR F (!! 0))
(*DEFVAR G (!! 0))

(*PROCLAIM '(type (field-pvar TY-SZ) NX ATY BTY LTY LLTY))

(*DEFVAR NX (!! 0))
(*DEFVAR ATY (!! 0))
(*DEFVAR BTY (!! 0))
(*DEFVAR LTY (!! 0))
(*DEFVAR LLTY (!! 0))

; Codes for next operation, not all are used as such

(defmacro cPNT!! () '(!! 1))       (defmacro cCHK!! () '(!! 2))
;; (defmacro cOR!!  () '(!! 3))       (defmacro cUNI!! () '(!! 4))
;; (defmacro cVAL!! () '(!! 5))       (defmacro cAND!! () '(!! 6))
(defmacro cCXT!! () '(!! 7))       (defmacro cSPL!! () '(!! 8))
(defmacro cNM!!  () '(!! 9))       (defmacro c5NW!! () '(!! 10))

; Tests for next operation

(defmacro tPNT!! () '(=!! NX (!! 1)))   (defmacro tCHK!! () '(=!! NX (!! 2)))
;; (defmacro tOR!!  () '(=!! NX (!! 3)))  (defmacro tUNI!! () '(=!! NX (!! 4)))
;; (defmacro tVAL!! () '(=!! NX (!! 5)))  (defmacro tAND!! () '(=!! NX (!! 6)))
(defmacro tCXT!! () '(=!! NX (!! 7)))   (defmacro tSPL!! () '(=!! NX (!! 8)))
(defmacro tNM!!  () '(=!! NX (!! 9)))   (defmacro t5NW!! () '(=!! NX (!! 10)))

; Working Booleans for FIRE and related programs

(*PROCLAIM '(TYPE BOOLEAN-PVAR MRKT CTVT CPYT SCPT LQNT))
(*DEFVAR MRKT nil!!)
(*DEFVAR CTVT nil!!)
(*DEFVAR CPYT nil!!)
(*DEFVAR SCPT nil!!)
(*DEFVAR LQNT nil!!)

; Used in reduca, reducb:

(defmacro activep (x)
"(activep x) -- x is an active pointer (not self or sysc)"
`(and!! (>!! ,x (!! syscmx)) (/=!! ,x (self-address!!)))
)

(defmacro atomic (x xty)
"(atomic x xty) -- x (of type xty) is sysc, symbol or number"
 `(or!! (<!! ,x (!! ATML)) (=!! ,xty (!! (ty-numb))))
)

(defmacro true-false (x)
"(true-false x) -- test for constants true, false"
 `(<=!! (logxor!! ,x (!! CB-false)) (!! 1))
)

; Address tests

(defmacro hgh!! (n) "(hgh!! n) -- n is potential CCPV node"
 `(>=!! ,n (!! ATML)))

(defmacro bit!! (f b) "Test bit b of field f" 
 `(plusp!! (load-byte!! ,f (!! ,b) (!! 1)))
)

; Widely used functions

(proclaim '(type function initv FIRE garbage report cleanup gatherstats wff))

(defvar MXCA 0 "Maximum cube-address+1")
(defvar SYMIS 0 "Name-code for IS")
(defvar REDERROR nil "For signalling error in reduction")

(defun setup ()
 "setup  -- initializes system with current cm array size"
 (setq MXCA  cm:*user-cube-address-limit*)
 (setq SYS-mode 'BUILDING)
 (initv)
 (*when (<=!! (self-address!!) (!! syscmx))
        (sG!! (!! (ty-dummy))) (sF!! (!! 1)))
 (mapc #'reset-typ oplst)
 (sS CB-[] T)  (sS CB-inf T)
 (setq SYMIS (X CB-is))
 (initstore)
;  Table for Booleans     false             true
 (setq and-ifcd (logior CB-false         (ash CB-I      4)   ; and
			(ash CB-I     8) (ash CB-true  12)   ; or
		        (ash CB-true 16) (ash CB-false 20)   ; not
		        (ash CB-K2   24) (ash CB-K     28))) ; if
 (*set AND-IF (!! and-ifcd))
 (setq ROOT 0) (+F ROOT)
 "Initialized"
)

(defun reset-typ (id) "Reset system constant id"
 (let ((p (get id 'OPERATOR))) (initty id (car p)))
)
